#/* Copyright 1998-2011 Research Foundation State University of New York */

#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

#!/usr/bin/python

import collections
import itertools
from numpy import *
import scipy
import scipy.linalg
import scipy.optimize
import traceback
import time
from qubx.fast.model import Eigen_Scipy_Ptr
from qubx.model_constraints import *

# begin_html
# <p>
# See also:
# <ul>
# <li><a href="max_subi_ll.h.html">max_subi_ll.h</a></li>
# <li><a href="max_subi_ll.cpp.html">max_subi_ll.cpp</a><li>
# <li><a href="max_inter_ll.py.html">max_inter_ll.py</a><li>
# </ul>
# <b>Up:</b> <a href="index.html">Index</a>
#</p>
# end_html

# ============= The Data =========================

# The data consist of one or more parallel signals.  The first (index 0) is the Markovian one to be analyzed.
# Each additional signal describes a ligand or voltage variable.
# A signal is a list of segments.  A segment of consists of "events" ("dwells").
# An event has duration in milliseconds, and class -- index of its measured amplitude.
# For each segment there's a list of class amplitude values.
# For each signal you provide a list of segments: [(classes[], durations[], amp_of_cls[]), ...]

# If the model depends on a variable which is held constant, provide a constant signal.
# This function copies the segmentation of a template signal:

def ConstantSignal(template, value):
    return [([0], [sum(durations)], [value]) for classes, durations, amps in template]

# We will multiplex the signals into a single idealized stream, whose classes are multi-classes (mcls):
#    mcls[i] = (model_class, stimulus_class)
# The stimulus classes (scls) are tuples with the value of each signal (signal 0 is ignored).

def MultiplexSignals(signals):
    """MultiplexSignals([(classes_0_seg0, durations_0_seg0, amp_of_cls_0_seg0), ...seg1...], ..._1_...)

Returns segments, multi_classes, stimulus_classes
where   segments         = [(mclasses_seg0, durations_seg0), (mclasses_seg1, durations_seg1), ...]
        multi_classes    = [(model_class_of_0, stimulus_class_of_0), (model_class_of_1, stimulus_class_of_1), ...],
        stimulus_classes = [(signal_0_scls_0, signal_1_scls_0, ...), (signal_0_scls_1, signal_1_scls_1, ...), ...]
"""
    # assuming each signal has the same segmentation, collect all the signals for each segment together
    stacked_segs = [ [seg] for seg in signals[0] ]
    nseg = len(stacked_segs)
    raw_segs = [segs for segs in signals if len(segs) == nseg]
    for segs in raw_segs[1:]:
        for j, seg in enumerate(segs):
            stacked_segs[j].append(seg)
            
    # We use the same multi- and stimulus-classes for all segments, numbered in
    # the order of appearance.  These dicts start counting unique indices from 0.
    # At the end we'll reverse them into lists, and the unique index will be the location.
    # sclass keys are tuples of float (amp_of_signal_i)
    sclass = collections.defaultdict(itertools.count().next)
    # mclass keys are pairs of integers (model_class, sclass)
    mclass = collections.defaultdict(itertools.count().next)

    def MplexEvents(stack):  # Called upon each stacked segment
        """MplexEvents([(classes_sig_0, durations_sig_0, amps_sig_0), ...])
        Finds multiplex and stimulus classes (populates sclass and mclass).
        Returns parallel arrays (multiclasses, durations)."""
        for i, cda in enumerate(stack):
            if (i>0) and (cda[0][0] < 0):
                raise Exception("Can't process events: one or more segments starts with a gap (excluded region).")
        try:
        N = [len(x[0]) for x in stack]                                  #  number of events on signal i
        finger = [0 for x in stack]                                     #  index of current event on signal i
        clss = [classes[0] for classes, durations, amp in stack]        #  class of current event on signal i
        amps = [amp[classes[0]] for classes, durations, amp in stack]   #  amp of current event on signal i
        remain = [durations[0] for classes, durations, amp in stack]    #  time remaining in event on signal i
        except IndexError:
            return [], []
        cl = []                            # outputs
        du = []
        while True: # all(f<n for f,n in itertools.izip(finger, N)):
            # loop until any signal ends
            done = False
            for i, n in enumerate(N):
                if finger[i] >= n:
                    done = True
            if done: break
            # find the signal with shortest remaining
            changing = argmin(remain)
            tm = remain[changing]
            # look up the current stimulus- and multi-classes
            sc = sclass[tuple([0]+amps[1:])]
            mc = mclass[(clss[0],sc)]
            # append the event
            cl.append(mc)
            du.append(tm)
            # shorten all signals' remain time, and move the finger for any that are 0
            for i, t in enumerate(remain):
                remain[i] = t - tm
                if remain[i] <= 0.0:
                    finger[i] += 1
                    if finger[i] < N[i]:
                        remain[i] = stack[i][1][finger[i]]
                        clss[i]   = stack[i][0][finger[i]]
                        if clss[i] >= 0:
                            amps[i]   = stack[i][2][clss[i]]
                        # else: gap: hold over last event's amp
        return cl, du
    mplex_segs = [MplexEvents(seg) for seg in stacked_segs]

    # convert the class dictionaries to lists:
    Nsc = len(sclass)
    amps_of_sc = [None] * Nsc
    for amps, sc in sclass.iteritems():
        amps_of_sc[sc] = amps
    Nmc = len(mclass)
    mclass_lst = [None]*len(mclass)
    for tup, mc in mclass.iteritems():
        mclass_lst[mc] = tup
    
    return mplex_segs, mclass_lst, amps_of_sc # [(classes, durations)] * nseg, [(I_cls, S_cls)] * ncls, {S_cls : {Name : amp}}


# Then we apply the dead time t, concatenating any too-short event into the previous one:

def ApplyDeadTime(classes, durations, tdead):
    """ApplyDeadTime(classes, durations, tdead)

    classes:      list of each event's class index
    durations:    list of each event's duration, in milliseconds
    tdead:        length of shortest perceptible event
    
    Returns: (classes, durations) with
                 * events shorter than tdead joined to the prior event
                 * remaining events shortened by tdead
                   (because the forward likelihood is calculated separately for
                    the first (duration-tdead) "dwell time" and the final
                    tdead of "transition time.")"""
    # copy the arrays and modify them in-place
    cl = array(classes, copy=True, dtype='int32')
    du = array(durations, copy=True)
    tm = 0.0
    L = len(du)
    # skip initial too-shorts
    first_alive = 0
    while first_alive < L:
        if du[first_alive] < tdead:
            tm += du[first_alive]
            first_alive += 1
        else:
            # du[first_alive] += tm # uncomment to join them to the first long-enough event
            break
    i_rd, i_wr = first_alive, 0
    while i_rd < L:
        cls, tm = cl[i_rd], du[i_rd]
        if i_wr and ((cl[i_wr-1] == cls) or (tm < tdead)):
            du[i_wr-1] += tm
        else:
            cl[i_wr] = cls
            du[i_wr] = tm - tdead
            i_wr += 1
        i_rd += 1
    cl.resize((i_wr))
    du.resize((i_wr))
    return cl, du

# the total data processing is to multiplex, apply dead time,
# and convert durations to seconds, for compatibility with Q-math

def ProcessSignals(tdead, signals):
    segs, mcls, scls = MultiplexSignals(signals)
    segs = [ApplyDeadTime(classes, durations, tdead) for classes, durations in segs]
    segs = [(classes, 1e-3*array(durations, dtype='float32')) for classes, durations in segs]
    return segs, mcls, scls


# ============= The Model ====================

# Our Markov model is a graph with colored vertices.  A vertex is called a "state,"
# and its color is a nonnegative integer called its "class."  States in the same
# class are indistinguishable (same amp).  To describe the vertices, provide the array
#    clazz = array([class of state s for s in len(states)])
#    Ns    = len(clazz)

# Each edge is labeled with its transition rate (probability per second).  These form the matrix
#    Q, a Ns x Ns matrix with
#    Q[a,b] = rate from state a to state b
#    Q[a,a] = - sum(Q[a,:])

# Each Q[a,b] is computed by q = k0 * ligand * e**(k1 * voltage).  You provide the Ns x Ns matrices
#    K0, K1   of kinetic parameters
#    L, V     index of the ligand or voltage signal influencing each rate, or 0
# with the diagonals undefined.

# A pair of states is either connected in both directions or neither.  To indicate un-connectedness, set
#    K0[a,b] = K0[b,a] = 0.0

def BuildQ(K0, K1, L, V, signal_values):
    Ns = K0.shape[0]
    Q = matrix(zeros(shape=K0.shape))
    for a in xrange(Ns):
        for b in xrange(Ns):
            if a != b:
                k = K1[a,b]
                if V[a,b]:
                    k *= signal_values[V[a,b]]
                k = exp(k)
                k *= K0[a,b]
                if L[a,b]:
                    k *= signal_values[L[a,b]]
                Q[a,b] = k
        Q[a,a] = - sum(Q[a])
    return Q


# For the initial probability vector we follow the lead of (Milescu ...) and
# recompute equilibrium probability each iteration.
# Actually, we'll use QtoPe(eQ) [eQ is the dead-time-corrected Q matrix, below].

def QtoPe(Q):
    # S = [Q | 1]; Pe = 1 * (S * S.T).I
    S = matrix(ones(shape=(Q.shape[0], Q.shape[1]+1)))
    S[:,:-1] = Q
    u = matrix(ones(shape=(1,Q.shape[0])))
    p = u * (S*S.T).I
    return array(p).reshape((Q.shape[0],))


# The core calculation is the log-likelihood (LL) of the model producing the data.
# We follow the method of (Qin 1996) but revise the missed event correction
# to allow the LL of one event to be computed piece-wise.  In MIL, the LL of a dwell
# and its subsequent transition are computed together, with no provision for a
# "transition" from a class to itself.  We re-use the corrected eQaa for the initial (t-td) of an event:

def QtoQe(Q, clazz, td): # (3,4b)
    # z = a-bar (not a)
    expm = scipy.linalg.matfuncs.expm
    eQ = matrix(zeros(shape=Q.shape))
    for a in set(clazz):
        # partitions:
        A = clazz == a
        Z = clazz != a
        Qaa = Q[ix_(A, A)]
        Qaz = Q[ix_(A, Z)]
        Qza = Q[ix_(Z, A)]
        Qzz = Q[ix_(Z, Z)]
        Iz = identity(Qzz.shape[0])
        # staying and returning:
        missed_returning = ((Qaz * (Iz - expm(td*Qzz))) * Qzz.I) * Qza
        eQ[ix_(A, A)] = Qaa - missed_returning
        
        for b in (set(clazz) - set([a])):
            # partitions
            B = clazz == b
            C = (clazz != a) * (clazz != b)
            Qab = Q[ix_(A, B)]
            Qbb = Q[ix_(B, B)]
            if any(C):
                Qac = Q[ix_(A, C)]
                Qcc = Q[ix_(C, C)]
                Qcb = Q[ix_(C, B)]
                Ic = identity(Qcc.shape[0])
            # switching class:
                escaped_indirect = ((Qac * (Ic - expm(td * Qcc))) * Qcc.I) * Qcb
            else:
                escaped_indirect = matrix(zeros(shape=Qab.shape))
            # as published:
            # eQ[ix_(A,B)] = expm(td * missed_returning) * (Qab - escaped_indirect)
            # as coded, since before 2000:
            eQ[ix_(A,B)] = (Qab - escaped_indirect) * expm(td * Qbb)
    return eQ

# Actually, we'll precompute part of exp(eQ*(t-td)) (7) so we can quickly change t:

def Spectrum(M):
    lamb, eig = linalg.eig(M)
    return matrix(eig), lamb, matrix(eig).I
def SpectrumExp(spectrum, t):
    U, lamb, Ui = spectrum
    return U*diag(exp(lamb*t))*Ui

# and compute the forward probability over the final td using the sampled transition matrix
#    A = e**(Q*td)
# with zeros for impossible transitions (into a different class than the data says).
# These are memorized to avoid duplicate calculations, with this general utility:

def memoize(func, lbl=""):
    """Returns wrapped func that caches return values for specific args.
       All args must be immutable (e.g. tuples not lists).

Example:
    def factorial(n):
        if n > 1:
           return n * fast_fac(n-1)
        else:
           return 1
     fast_fac = memoize(factorial)
"""
    results = {}
    def wrapper(*args):
        if args in results:
            return results[args]
        else:
            result = func(*args)
            results[args] = result
            # print lbl, args, ':', result
            return result
    return wrapper


def subi_ll_or_0(args): # note the only arg is a tuple for MSL args; this helps with multiprocessing.Pool.imap
    try:
        if args[0] and HAVE_LIB:
            return subi_ll_lib(*args[1:])
        else:
            return subi_ll(*args[1:])
    except KeyboardInterrupt:
        raise
    except:
        traceback.print_exc()
        return 987.654321


def subi_ll(clazz, P0, K0, K1, L, V, tdead_sec, segs, mcls, scls, printout=False):
    expm = scipy.linalg.matfuncs.expm
    td = tdead_sec
    
    if P0 != None:
        P0_norm = array(P0, copy=True)
        P0_norm /= sum(P0_norm) or 1.0
    
    if printout:
        print 'Nd:',[len(classes) for classes,durations in segs]
        
    # Q is different for each scls:
    QQ = [BuildQ(K0, K1, L, V, amps) for amps in scls]
    if printout:
        print 'Q:',QQ[-1]
    try:
        eQQ = [QtoQe(Q, clazz, td) for Q in QQ] # (3,4b)
        if printout:
            print 'eQ:',eQQ[-1]
    except linalg.linalg.LinAlgError, err:
        print err,'; falling back to raw Q'
        eQQ = QQ
    # the submatrices eQaa are used for dwell probability (3)
    eQQaa = [ [(a, eQ[ix_(clazz==a, clazz==a)]) for a in set(clazz)]
              for eQ in eQQ ]
    # we take their spectral decomposition for quick exponentiation:
    eQQaaSpectrum = [ dict([(a, Spectrum(eQaa)) for a, eQaa in Qaas])
                      for Qaas in eQQaa ]
    # class < 0 is a gap
    eQQgapSpectrum = [Spectrum(eQ) for eQ in eQQ]
    if printout:
        print 'eQcc spectrum:'
        print eQQaaSpectrum[-1][0]
    def DwellProbMatrix(multiclass, t): # (7)
        G_full = matrix(zeros(shape=(len(clazz),len(clazz))))
        mc, sc = mcls[multiclass]
        if mc >= 0:
            A = clazz == mc
            U, lamb, Ui = eQQaaSpectrum[sc][mc]
            G_full[ix_(A,A)] = U*diag(exp(lamb*t))*Ui
        else: # gap: exponentiate full eQ matrix
            U, lamb, Ui = eQQgapSpectrum[sc]
            G_full[:,:] = U*diag(exp(lamb*t))*Ui
        return G_full
    
    # for each scls's Q, there's an A(td):
    AA = [expm(Q*td) for Q in QQ] # (8)
    if printout:
        print 'Atd:', AA[-1]
    # and we zero out different columns depending on the next event.
    # To specify 'any class but a', give destination=-(a+1)
    # destination=None: don't zero anything
    def _Adead(sc, destination=None): # (9)
        if destination == None:
            return AA[sc]  # gap: full A matrix
        else:
        Adead = matrix(AA[sc], copy=True)
        if destination >= 0:
            cols = clazz == destination
        else:
            cols = clazz != - (destination + 1)
        all_rows = [True] * len(clazz)
        Adead[ix_(all_rows, ~cols)] = 0.0
        return Adead
    # presumably the same transitions happen again and again, so we compute and memorize on first need:
    Adead = memoize(_Adead, 'Adead') # (9)
    
    # begin_html
    # \(LL = log(\alpha_N 1)\), where \(\alpha_k\) is the column vector of state probabilities at time k.
    # As in (Qin ...), we find \(scale_k = \frac{1}{\sum \alpha_k}\) and reset each \(\alpha_k = \alpha_k * scale_k\), so that
    # \[LL = - \sum_k log(scale_k)\]
    # end_html
    def Scale(ak):
        mag = sum(ak)
        if mag == 0.0:
            raise Exception('Impossible!')
        scale = 1.0 / mag
        ak *= scale
        return scale
    
    LL = 0.0 # sum of segment LL

    for classes, durations in segs:
        Nd = len(durations)
        scale = zeros(shape=(Nd+1,))
        #alpha = matrix(zeros(shape=(Nd+1, len(clazz))))
        ak = matrix(zeros(shape=(1, len(clazz))))
        
        mc = classes[0]
        b, sc = mcls[mc]
        if P0 != None:
            P0_seg = array(P0, copy=True)
        else:
            P0_seg = QtoPe(eQQ[sc])
        if b >= 0: #no gap on start: zero impossible states
            P0_seg *= (clazz == b)
        # P0 /= sum(P0) or 1.0 ??
        ak[0,:] = P0_seg
        scale[0] = Scale(ak)
        #alpha[0] = ak

        for k in xrange(Nd-1):
            ak *= DwellProbMatrix(mc, durations[k]) # (7) modified for gaps

            mc = classes[k+1]
            a, b = b, mcls[mc][0]
            if b >= 0:
                ak *= Adead(sc, b) # (9)
            else: # transition into gap: full Adead matrix
                ak *= Adead(sc, None)
            scale[k+1] = Scale(ak) # (10)
            #alpha[k+1] = ak

            sc = mcls[mc][1]
        a = b
        ak *= DwellProbMatrix(mc, durations[Nd-1])
        if a >= 0:
            ak *= Adead(sc, -(a+1))
        else: # end in gap; full Adead
            ak *= Adead(sc, None)
        scale[Nd] = Scale(ak)
        #alpha[Nd] = ak

        LL -= sum(log(scale)) # (10)
    return LL


try:
    import ctypes
    try:
        maxill = ctypes.windll.LoadLibrary('maxill.dll')
    except:
        maxill = ctypes.cdll.LoadLibrary('libmaxill.so')
    
    def subi_ll_lib(clazz, P0, K0, K1, L, V, tdead_sec, segs, mcls, scls, printout=False):
        p_int = ctypes.POINTER(ctypes.c_int)
        p_float = ctypes.POINTER(ctypes.c_float)
        p_double = ctypes.POINTER(ctypes.c_double)

        clazz_ = array(clazz, dtype='int32')
        Nseg = len(segs)
        if P0 != None:
            P0_ = array(P0)
        else:
            P0_ = None
        K0_ = array(K0)
        K1_ = array(K1)
        L_ = array(L)
        V_ = array(V)
        dwellCounts = array([len(classes) for classes, durations in segs], dtype='int32')
        classeses = (p_int*Nseg)(*[classes.ctypes.data_as(p_int) for classes, durations in segs])
        durationses = (p_float*Nseg)(*[durations.ctypes.data_as(p_float) for classes, durations in segs])
        Nsig = len(scls[0])
        Nplex = len(mcls)
        plexicls = zeros(dtype='int32', shape=(2*Nplex))
        for i,mc in enumerate(mcls):
            plexicls[2*i  ], plexicls[2*i+1] = mc
        Nstim = len(scls)
        stimcls = zeros(shape=(Nstim, Nsig))
        for i,sc in enumerate(scls):
            stimcls[i,:] = sc
        ll = array([-9898.0])
        
        def cdata(x, typ):
            if x == None: return x
            return x.ctypes.data_as(typ)
        
        rtnval = maxill.subi_ll_arr(len(clazz), cdata(clazz_, p_int), cdata(P0_, p_double),
                                    cdata(K0_, p_double), cdata(K1_, p_double), cdata(L_, p_int), cdata(V_, p_int),
                                    ctypes.c_double(tdead_sec), Nseg, cdata(dwellCounts, p_int), classeses, durationses,
                                    Nsig, Nplex, cdata(plexicls, p_int), Nstim, cdata(stimcls, p_double),
                                    cdata(ll, p_double), Eigen_Scipy_Ptr)
        return ll[0]
    HAVE_LIB = True
except KeyboardInterrupt:
    raise
except:
    print 'Compiled maxill library was not found.'
    HAVE_LIB = False


# What are the optimization parameters?  (some of) the off-diagonal elements of K0 and K1.
# Actually, we use log(K0) to keep them positive.
# Breaking with past practice, we include all the unconnected zeros,
# and put all the K0s before all the K1s.  As we'll see, this is a column vector.

def K_index(N):
    ix = []
    if N <= 0: return ix
    for a in xrange(N):
        for b in xrange(N):
            if a != b:
                ix.append((a,b))
    return ix

def K01toK(K0, K1):
    k0, k1 = [], []
    for a,b in K_index(K0.shape[1]):
        if a != b:
            x = K0[a,b]
            k0.append((x > 0.0) and log(x) or 0.0) # tricky: 1->0 and 0->0; see KtoK01
            k1.append(K1[a,b])
    K = matrix(zeros(shape=(len(k0)+len(k1), 1)))
    K[:,0] = matrix(array(k0 + k1)).T
    return K

# K alone isn't enough to reconstruct K0 and K1, since 0 could mean log(1), or else unconnected.
# We disambiguate by checking the original K0 for zeros, which mean unconnected.

def KtoK01(K, last_K0):
    # len(K)/2 = K0.N * (K0.N - 1);  K0.N = (1 + sqrt(1 + 2*len(K))) / 2.0
    N = int(round((1 + sqrt(1 + 2*len(K))) / 2.0))
    ixs = K_index(N)
    K0 = matrix(zeros(shape=(N,N)))
    i = 0
    for a,b in ixs:
        if a != b:
            if last_K0[a,b]:
                K0[a,b] = exp(K[i,0])
            i += 1
    K1 = matrix(zeros(shape=(N,N)))
    for a,b in ixs:
        if a != b:
            K1[a,b] = K[i]
            i += 1
    return K0, K1


# But some of those K are zero and have to stay zero, and the user may have other constraints.
# We represent the constraints by the system of linear equations (11) and derive the transformations (12, 13).

# Some constraints are more a part of the representation than the model:
#   * where K0 == 0 there's no connection; fix k0 and k1
#   * where V == 0  fix k1
def AutoConstraints(K0, K1, L, V):
    """Returns the constraints (Ain, Bin)   [11] which fix unconnected k0 and unsensitive k1."""
    aa = []
    bb = []
    ixs = K_index(K0.shape[0])
    for i, xx in enumerate(ixs):
        a, b = xx
        if not K0[a,b]:
            coeffs = zeros(shape=(1,2*len(ixs)))
            coeffs[0, i] = 1.0
            aa.append(coeffs)
            bb.append(0.0)
            V[a,b] = 0
        if not V[a,b]:
            coeffs = zeros(shape=(1,2*len(ixs)))
            coeffs[0, len(ixs)+i] = 1.0
            aa.append(coeffs)
            bb.append(0.0)
    if aa:
        return vstack(aa), matrix(array(bb)).T
    else:
        return matrix(zeros(shape=(0,0))), matrix(zeros(shape=(0,1)))

# The user might also have some constraints for our linear system:
def UserConstraints(K0, K1, L, V, constraints):
    """Returns the constraints (Ain, Bin)   [11]
    corresponding to the tuples    (name, states) in constraints
    where name is in ["FixRate", "FixExp", "ScaleRate", "ScaleExp", "BalanceLoop", "ImbalanceLoop"]
    and states is a sequence of state indices.  The first four names require 2,2,4,4 states respectively;
    the last two require at least 3."""
    aa = []
    bb = []
    ixs = K_index(K0.shape[0])
    ixo = dict([(ab, i) for i, ab in enumerate(ixs)])
    N = len(ixs)
    for name, states  in constraints:
        if name == 'FixRate':
            r = ixo[(states[0], states[1])]
            coeffs = zeros(shape=(1,2*N))
            coeffs[0, r] = 1.0
            aa.append(coeffs)
            bb.append(log(K0[states[0], states[1]]))
        if name == 'FixExp':
            r = ixo[(states[0], states[1])]
            coeffs = zeros(shape=(1,2*N))
            coeffs[0, N+r] = 1.0
            aa.append(coeffs)
            bb.append(K1[states[0], states[1]])
        if name == 'ScaleRate':
            r1 = ixo[(states[0], states[1])]
            r2 = ixo[(states[2], states[3])]
            coeffs = zeros(shape=(1,2*N))
            coeffs[0, r1] = 1.0
            coeffs[0, r2] = -1.0
            aa.append(coeffs)
            bb.append(log(K0[states[0], states[1]]) - log(K0[states[2], states[3]]))
        if name == 'ScaleExp':
            r1 = ixo[(states[0], states[1])]
            r2 = ixo[(states[2], states[3])]
            coeffs = zeros(shape=(1,2*N))
            coeffs[0, N+r1] = 1.0
            coeffs[0, N+r2] = - K1[states[0], states[1]] / K1[states[2], states[3]]
            aa.append(coeffs)
            bb.append(0.0)
        elif name == 'LoopBal':
            v_depend = False
            loopst = list(states) + [states[0]]
            coeffs = zeros(shape=(1,2*N))
            for i in xrange(len(loopst)-1):
                rf, rb = rate_ix[(loopst[i], loopst[i+1])], rate_ix[(loopst[i+1], loopst[i])]
                coeffs[0,rf] = 1.0
                coeffs[0,rb] = -1.0
                v_depend = v_depend or V[(loopst[i], loopst[i+1])] or V[(loopst[i+1], loopst[i])]
            aa.append(coeffs)
            bb.append(0.0)
            if v_depend:
                coeffs = zeros(shape=(1,2*N))
                for i in xrange(len(loopst)-1):
                    rf, rb = rate_ix[(loopst[i], loopst[i+1])], rate_ix[(loopst[i+1], loopst[i])]
                    coeffs[0,N+rf] = 1.0
                    coeffs[0,N+rb] = -1.0
                aa.append(coeffs)
                bb.append(0.0)
        elif typ == 'LoopImbal':
            v_depend = False
            loopst = list(states) + [states[0]]
            coeffs = zeros(shape=(1,2*N))
            b = 0.0
            for i in xrange(len(loopst)-1):
                rf, rb = rate_ix[(loopst[i], loopst[i+1])], rate_ix[(loopst[i+1], loopst[i])]
                coeffs[0,rf] = 1.0
                coeffs[0,rb] = -1.0
                b += log(K0[(loopst[i], loopst[i+1])]) - log(K0[(loopst[i+1], loopst[i])])
                v_depend = v_depend or V[(loopst[i], loopst[i+1])] or V[(loopst[i+1], loopst[i])]
            aa.append(coeffs)
            bb.append(b)
            if v_depend:
                coeffs = zeros(shape=(1,2*N))
                for i in xrange(len(loopst)-1):
                    rf, rb = rate_ix[(loopst[i], loopst[i+1])], rate_ix[(loopst[i+1], loopst[i])]
                    coeffs[0,N+rf] = 1.0
                    coeffs[0,N+rb] = -1.0
                aa.append(coeffs)
                bb.append(0.0)

    if aa:
        return vstack(aa), matrix(array(bb)).T
    else:
        return matrix(zeros(shape=(0,N))), matrix(zeros(shape=(0,1)))


# You could think of others too; just vstack them with the rest:
def AllConstraints(K0, K1, L, V, constraints):
    """Returns combined AutoConstraints(...) and UserConstraints(...)"""
    A, B = UserConstraints(K0, K1, L, V, constraints)
    Aauto, Bauto = AutoConstraints(K0, K1, L, V)
    if A.shape[0] == 0:
        return Aauto, Bauto
    elif Aauto.shape[0] == 0:
        return A, B
    else:
        return vstack((A, Aauto)), vstack((B, Bauto))

# To derive the transformation we will use singular value decomposition (svd)
# which is really well described at http://www.uwlax.edu/faculty/will/svd/index.html

# This wrapper around numpy.linalg.svd makes it behave more like LAPACK:
#   * V is transposed
#   * matrices have full dimension
# Then we repair and check it:
#   * tiny numbers are replaced with 0
#   * U and V need orthonormal columns
# And derive some useful extras:
#   * Winv, where inv(0) = 0
#   * Ainv = V*Winv*U.T
#   * Nsv = number of nonzero singular values
def adjusted_svd(A):
    u,s,v = linalg.svd(A, full_matrices=True)
    N = s.shape[0]
    U = matrix(zeros(shape=(max(A.shape[0], u.shape[0]), max(A.shape[1], u.shape[1]))))
    U[:u.shape[0],:u.shape[1]] = u
    S = zeros(shape=(U.shape[1],))
    S[:N] = s
    N = len(s[s>=1e-10])
    W = diag(S)
    V = matrix(zeros(shape=(A.shape[1], W.shape[1])))
    V[:v.shape[1],:v.shape[0]] = v.T 
    for M in U, W, V:
        for i in xrange(M.shape[0]):
            for j in xrange(M.shape[1]):
                if abs(M[i,j]) < 1e-10:
                    M[i,j] = 0.0
    for i,x in enumerate(S):
        if abs(x) < 1e-10:
            S[i] = 0.0
    Sinv = array([x and 1/x or 0 for x in S])
    Winv = diag(Sinv)
    # A = U*W*V.T; A.I = V*Winv*U.T
    if linalg.norm(identity(U.shape[0]) - U*U.T) > 1e-8:
        raise Exception("Bad SVD: U's columns are not orthonormal")
    if linalg.norm(identity(V.shape[0]) - V*V.T) > 1e-8:
        raise Exception("Bad SVD: V's columns are not orthonormal")
    return U, W, V, Winv, V*(Winv*U.T), N

# This function cleans up the constraint system prior to the transform,
# by deleting duplicate constraints.  It uses two checks in case of numerical inaccuracy.
# It's called automatically in linear_constraints().
def reduce_constraints(A, B):
    Csrc = matrix(hstack((A, B)).T) # may not be square
    N = max(Csrc.shape)
    C = matrix(zeros(shape=(N,N))) # (14)
    C[:Csrc.shape[0],:Csrc.shape[1]] = Csrc
    Csub = C
    cols = [C[:,i] for i in xrange(Csrc.shape[1])]
    i = 0
    rank = 0
    while i < len(cols):
        Csub = hstack(cols[:i+1]) # (15)
        U, W, V, Winv, Cinv, Nsv = adjusted_svd(Csub) # (16)
        if Nsv == rank: # check 1: added constraint must increase rank (21)
            del cols[i]
        else:
            rank = Nsv
            j = i+1
            while j < len(cols):
                # check 2: other constraints can't be lin. combinations of the prior
                image = Cinv * cols[j] # (19)
                if 1e-5 > linalg.norm(cols[j] - (Csub*image)):
                    del cols[j]
                else:
                    j += 1
            i += 1
    return matrix(Csub[:-1,:].T), matrix(Csub[-1,:].T)
  

# Now that we have K0, K1, L, V and rate vector K,  
# we do some magic with the SVD to change your linear constraints
#    A * K = B
# into a transformation between K and the column vector P of free parameters
#    K = A*P + B
#    P = Ainv * K

def linear_constraints(Ain, Bin, K0):
    """Returns A, B, A.I, P0 [12]."""
    A, B = reduce_constraints(Ain, Bin)
    Nc = A.shape[0]
    Nk = A.shape[1]
    if A.shape[0] == 0:
        return identity(len(K0)), matrix(zeros(shape=(len(K0),1))), matrix(array([x for x in K0])).T
    U, W, V, Winv, Ainv, rank = adjusted_svd(A)
    if rank < Nc:
        raise Exception('Bad SVD: Some constraints may be incompatible')
    if rank != Nc:
        raise Exception('Too many constraints.  %i constraints may be linear combinations of the others.' % (Nc - rank))
    for i in xrange(Nc):
        for j in xrange(Nc,Nk):
            if abs(U[i,j] > 1e-5):
                raise Exception('Bad SVD: 0 != U[c,non] == %.6g' % U.A[i,j])
    Acns = matrix(V[:,Nc:]) # (24)
    Bcns = Ainv * B         # (29)
    Out0 = Acns.T*K0        # (13)
    return Acns, Bcns, Acns.T, Out0

# And yet another linear transformation, this time to scale initial parameter values to 1.0.
# (In the case where p[i] == 0, we scale by 1 and translate by -1.)
# P = As*S + Bs
# S = As.I * (P - Bs)

def start_at_ones(P0):
    """Returns As, Bs, As.I, S0 [32]."""
    scales = array([x or 1.0 for x in P0.A[:,0]])
    reverses = array([x and 1/x or 1.0 for i,x in enumerate(P0.A[:,0])])
    Ascal = matrix(diag(scales))          # (35)
    Bscal = matrix(zeros(shape=P0.shape)) # (36)
    for i,x in enumerate(P0.A[:,0]):
        if not x:
            Bscal[i] = -1.0
    return Ascal, Bscal, matrix(diag(reverses)), matrix(ones(P0.shape))

def KToPars(rates, A, Ainv, B): # (13)
    k = matrix(zeros(shape=(prod(rates.shape),1)))
    k[:,0] = matrix(rates.flatten()).T
    return Ainv*k

# variant 1: pars is a python or nd- array, such as S provided by an optimizer
def parsToK(pars, A, Ainv, B): # (12, 32)
    pa = array(pars).flatten()
    P = matrix(pa).T
    K = ParsToK(P, A, Ainv, B)
    return K

# variant 2: pars is a column vector, such as P = parsToK(S, ...)
def ParsToK(P, A, Ainv, B): # (12, 32)
    return A*P+B


# To set up, you provide K0,K1,L,V and constraints, and we derive
# P and S and their transforms.  Then the optimizer adjusts S and we use it
# to build the Q matrices:
#   S --parsToK---> P --ParsToK---> K --KtoK01---> K0,K1 --BuildQ---> Q



# Here we use simplex to maximize likelihood of user-supplied or stock data and model.
# TODO:  standardize output for diff-testing
#        explain


TDEAD = 0.1 # the sampling rate since the simulator is perfect
MODELFILE = 'msl_test.qmf'
DATAFILE = 'msl_test.qsf'

if __name__ == '__main__':
    import sys
    import qubx.model
    import qubx.tree
    qubx.tree.CHOOSE_FLAVOR('numpy')

    if '--help' in sys.argv:
        print """usage: max_subi_ll.py [<tdead=0 in ms> [<model>.qmf [<data>.qsf]]]
or     max_subi_ll.py --verify"""
        sys.exit(0)
    if '--verify' in sys.argv:
        import subprocess
        p = subprocess.Popen(sys.argv[0] + " | diff msl_test.out -", shell=True)
        p.communicate()
        sys.exit(0)

    def QSFtoSignals(session, wanted_signals=None):
        raw_idls = [(i, str(dc['Name'].data), chan)
                    for i, dc, chan in itertools.izip(itertools.count(),
                                                      qubx.tree.children(session['DataChannels'], 'Channel'),
                                                      qubx.tree.children(session['Idealization'], 'Channel'))
                    if not chan.find('Segment').isNull]
        wanted_idls = [(nm, chan) for i, nm, chan in raw_idls if ((i==0) or (wanted_signals==None) or (nm in wanted_signals))]
        raw_segs = [(nm, [(seg['Classes'].storage.data[:,0],
                           seg['Durations'].storage.data[:,0],
                           seg['amp'].storage.data[:,0])
                          for seg in qubx.tree.children(chan, 'Segment')])
                    for nm, chan in wanted_idls]
        return [nm for nm, segs in raw_segs], [segs for nm, segs in raw_segs]
        # [name_0, ...],. [(classes_0_seg0, durations_0_seg0, amp_of_cls_0_seg0), ...seg1...]

    tdead = (len(sys.argv) > 1) and float(sys.argv[1]) or TDEAD
    print 'Tdead:', tdead

    model_path = (len(sys.argv) > 2) and sys.argv[2] or MODELFILE
    data_path = (len(sys.argv) > 3) and sys.argv[3] or DATAFILE

    model = qubx.model.OpenQubModel(model_path)
    print model
    rates = model.rates
    wanted_signals = set([rates.get(r, 'Ligand') for r in xrange(rates.size)] +
                         [rates.get(r, 'Voltage') for r in xrange(rates.size)])

    qsf = qubx.tree.Open(data_path, True)
    signal_names, signals = QSFtoSignals(qsf, wanted_signals)
    segments, multi_classes, stimulus_classes = ProcessSignals(tdead, signals)
    print 'Events:', sum(len(cl) for cl,du in segments)

    td_sec = tdead * 1e-3
    clazz = qubx.model.ModelToClazz(model)
    K0, K1, L, V = qubx.model.ModelToRateMatrices(model, signal_names)
    K = K01toK(K0, K1)
    
    t0 = time.time()
    LL = subi_ll(clazz, None, K0, K1, L, V, td_sec, segments, multi_classes, stimulus_classes, printout=True)
    print 'Initial LL:', LL, '[', ('%.3f' % (time.time() - t0)), 'sec ]'

    if HAVE_LIB:
        t0 = time.time()
        LL = subi_ll_lib(clazz, None, K0, K1, L, V, td_sec, segments, multi_classes, stimulus_classes, printout=True)
        print 'Compiled LL:', LL, '[', ('%.3f' % (time.time() - t0)), 'sec ]'

    constraints = qubx.model.ModelToConstraints(model)
    Asys, Bsys = AllConstraints(K0, K1, L, V, constraints)
    print 'Constraints in:'
    print Asys
    print Bsys
    Acns, Bcns, Ainv, pars = linear_constraints(Asys, Bsys, K)
    print 'Constraints out:'
    print Acns
    print Bcns
    Ascal, Bscal, Asci, pars = start_at_ones(pars)
    if any(pars != 1.0):
        print 'Non-unit starting par!',pars
        print Ascal
        print Bscal
        print Asci

    msl_ll = HAVE_LIB and subi_ll_lib or subi_ll
    def msl_func(pars):
        P = parsToK(pars, Ascal, Asci, Bscal)
        K = ParsToK(P, Acns, Ainv, Bcns)
        new_K0, new_K1 = KtoK01(K, K0)
        LL = msl_ll(clazz, None, new_K0, new_K1, L, V, td_sec, segments, multi_classes, stimulus_classes)
        print LL,pars
        return - LL
    counter = itertools.count()
    def msl_iter(pars):
        print '---> %i <---'%counter.next(), pars
        P = parsToK(pars, Ascal, Asci, Bscal)
        K = parsToK(P, Acns, Ainv, Bcns)
        print 'K',K.T
        new_K0, new_K1 = KtoK01(K, K0)
        qubx.model.K01ToModel(new_K0, new_K1, model)
        print model.rates
        print '------------->    <---'
    
    pars, fopt, iter, funcalls, warnflag = \
        scipy.optimize.fmin(msl_func, pars, maxfun=2000,
                            xtol=.0001, ftol=.0001, maxiter=300,
                            full_output=1, disp=1, retall=0, callback=msl_iter)
    print 'final',
    msl_iter(pars)
